# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module for writing data with SqlAlchemy."""

from __future__ import annotations

try:
  import sqlalchemy
except ImportError as e:
  raise ImportError(
    'Please install garf-io with sqlalchemy support '
    '- `pip install garf-io[sqlalchemy]`'
  ) from e

import logging

import pandas as pd
from garf.core import report as garf_report
from garf.io import formatter
from garf.io.telemetry import tracer
from garf.io.writers import abs_writer

logger = logging.getLogger(__name__)


class SqlAlchemyWriter(abs_writer.AbsWriter):
  """Handles writing GarfReports data to databases supported by SqlAlchemy.

  Attributes:
      connection_string:
          Connection string to database.
          More at https://docs.sqlalchemy.org/en/20/core/engines.html.
      if_exists:
          Behaviour when data already exists in the table.
  """

  def __init__(
    self, connection_string: str, if_exists: str = 'replace', **kwargs
  ):
    """Initializes SqlAlchemyWriter based on connection_string.

    Args:
        connection_string: Connection string to database.
    if_exists: Behaviour when data already exists in the table.
    """
    super().__init__(**kwargs)
    self.connection_string = connection_string
    self.if_exists = if_exists

  @tracer.start_as_current_span('sqldb.write')
  def write(self, report: garf_report.GarfReport, destination: str) -> None:
    """Writes Garf report to the table.

    Args:
        report: GarfReport to be written.
        destination: Name of the output table.
    """
    report = self.format_for_write(report)
    destination = formatter.format_extension(destination)
    dtypes = {}
    for column in report.column_names:
      if (report and isinstance(report[0][column], dict)) or (
        not report and isinstance(report.results_placeholder[0][column], dict)
      ):
        dtypes.update({column: sqlalchemy.types.JSON})
    if not report:
      df = pd.DataFrame(
        data=report.results_placeholder, columns=report.column_names
      ).head(0)
    else:
      df = report.to_pandas()
    logger.debug('Writing %d rows of data to %s', len(df), destination)
    write_params = {
      'name': destination,
      'con': self.engine,
      'index': False,
      'if_exists': self.if_exists,
    }
    if dtypes:
      write_params.update({'dtype': dtypes})
    df.to_sql(**write_params)
    logger.debug('Writing to %s is completed', destination)

  @property
  def engine(self) -> sqlalchemy.engine.Engine:
    """Creates engine based on connection string."""
    return sqlalchemy.create_engine(self.connection_string)



# Copyright 2026 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Writes GarfReport to Google PubSub topic."""

import json
import logging
import os

from garf.core import report as garf_report
from garf.io.telemetry import tracer
from garf.io.writers import abs_writer

try:
  from google.cloud import pubsub_v1
except ImportError as e:
  raise ImportError(
    'Please install garf-io with PubSub support - `pip install garf-io[pubsub]`'
  ) from e

logger = logging.getLogger(__name__)


class PubSubWriter(abs_writer.AbsWriter):
  """Publishes Garf Report to a pubsub topic.

  Attributes:
    topic_id: Id of PubSub topic.
  """

  def __init__(
    self,
    project: str = os.getenv('GOOGLE_CLOUD_PROJECT'),
    **kwargs: str,
  ) -> None:
    """Initializes PubSubWriter based on project."""
    super().__init__(**kwargs)
    self.project = project
    self._publisher = None

  @property
  def publisher(self) -> pubsub_v1.PublisherClient:
    if not self._publisher:
      self._publisher = pubsub_v1.PublisherClient()
    return self._publisher

  @tracer.start_as_current_span('pubsub.write')
  def write(self, report: garf_report.GarfReport, destination: str) -> None:
    topic_path = self.publisher.topic_path(self.project, destination)
    if not self.publisher.get_topic(request={'topic': topic_path}):
      self.publisher.create_topic(request={'name': topic_path})
    future = self.publisher.publish(
      topic_path, json.dumps(report.to_list('dict')).encode('utf-8')
    )
    return f'[Pubsub] - published message ID {future.result()}'



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc

import smart_open
from garf.io.telemetry import tracer
from typing_extensions import override


class AbsReader(abc.ABC):
  """Base class for reading queries."""

  @abc.abstractmethod
  def read(self, query_path: str, **kwargs: str):
    """Reads query from local or remote storage."""


class FileReader(AbsReader):
  """Reads from file."""

  @override
  @tracer.start_as_current_span('file.read')
  def read(self, query_path, **kwargs):
    with smart_open.open(query_path, 'r') as f:
      return f.read()


class ConsoleReader(AbsReader):
  """Read query from standard input."""

  @override
  @tracer.start_as_current_span('console.read')
  def read(self, query_path, **kwargs):
    return query_path


class NullReader(AbsReader):
  """Invalid reader."""

  def __init__(self, reader_option, **kwargs):
    raise ValueError(f'{reader_option} is unknown reader type!')

  def read(self):
    print('Unknown reader type!')


_READER_OPTIONS = {
  'file': FileReader,
  'console': ConsoleReader,
}


def create_reader(reader_option: str, **kwargs) -> AbsReader:
  if reader := _READER_OPTIONS.get(reader_option):
    return reader(**kwargs)
  return NullReader(reader_option)



# Copyright 2026 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Writes GarfReport to Kafka topic."""

import json
import logging

from garf.core import report as garf_report
from garf.io.telemetry import tracer
from garf.io.writers import abs_writer

try:
  from kafka import KafkaProducer
except ImportError as e:
  raise ImportError(
    'Please install garf-io with Kafka support - `pip install garf-io[kafka]`'
  ) from e

logger = logging.getLogger(__name__)
logging.getLogger('kafka.conn').setLevel(logging.WARNING)


class KafkaWriter(abs_writer.AbsWriter):
  """Publishes Garf Report to a Kafka topic.

  Attributes:
    bootstrap_servers: Kafka bootstrap servers.
  """

  def __init__(
    self,
    bootstrap_servers: str = 'localhost:9092',
    **kwargs: str,
  ) -> None:
    """Initializes KafkaWriter."""
    super().__init__(**kwargs)
    if isinstance(bootstrap_servers, str):
      bootstrap_servers = bootstrap_servers.split(',')
    self.bootstrap_servers = bootstrap_servers
    self._producer = None

  @property
  def producer(self) -> KafkaProducer:
    if not self._producer:
      self._producer = KafkaProducer(bootstrap_servers=self.bootstrap_servers)
    return self._producer

  @tracer.start_as_current_span('kafka.write')
  def write(self, report: garf_report.GarfReport, destination: str) -> str:
    """Writes report to Kafka topic.

    Args:
      report: GarfReport to write.
      destination: Kafka topic name.
    """
    future = self.producer.send(
      topic=destination,
      value=json.dumps(report.to_list('dict')).encode('utf-8'),
    )
    result = future.get(timeout=60)
    return (
      f'[Kafka] - published message to {result.topic} '
      'partition {result.partition} offset {result.offset}'
    )



# Copyright 2026 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Writes GarfReport to Elasticsearch index."""

import logging
from typing import Any, List, Union

from garf.core import report as garf_report
from garf.io.telemetry import tracer
from garf.io.writers import abs_writer

try:
  from elasticsearch import Elasticsearch, helpers
except ImportError as e:
  raise ImportError(
    'Please install garf-io with Elasticsearch support - '
    '`pip install garf-io[elasticsearch]`'
  ) from e

logger = logging.getLogger(__name__)


class ElasticsearchWriter(abs_writer.AbsWriter):
  """Writes Garf Report to Elasticsearch.

  Attributes:
    client: Elasticsearch client.
  """

  def __init__(
    self,
    hosts: Union[List[str], str] = 'localhost:9200',
    http_auth: Any = None,
    **kwargs: Any,
  ) -> None:
    """Initializes ElasticsearchWriter.

    Args:
      hosts: Elasticsearch hosts.
      http_auth: Authentication credentials (user, password) or similar.
      **kwargs: Other arguments for Elasticsearch client.
    """
    super().__init__(**kwargs)
    if isinstance(hosts, str):
      hosts = hosts.split(',')

    self.client = Elasticsearch(hosts=hosts, http_auth=http_auth, **kwargs)

  def _create_index_if_not_exists(self, index_name: str) -> None:
    """Creates index if it does not exist."""
    if not self.client.indices.exists(index=index_name):
      self.client.indices.create(index=index_name)
      logger.info(f'Created index {index_name}')

  @tracer.start_as_current_span('elasticsearch.write')
  def write(self, report: garf_report.GarfReport, destination: str) -> str:
    """Writes report to Elasticsearch.

    Args:
      report: GarfReport to write.
      destination: Elasticsearch index name.
    """
    self._create_index_if_not_exists(destination)

    data = report.to_list(row_type='dict')
    actions = [{'_index': destination, '_source': row} for row in data]

    success, failed = helpers.bulk(self.client, actions)
    return (
      f'[Elasticsearch] - successfully indexed {success} documents to '
      f'{destination}. Failed: {failed}'
    )



# Copyright 2026 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Writes GarfReport to Excel.

Writing to remote storage systems (gs, s3, hadoop, stfp) is also supported.
"""

from __future__ import annotations

import logging
import os
import pathlib
from typing import Optional, Union

import smart_open
from garf.core import report as garf_report
from garf.io import formatter
from garf.io.telemetry import tracer
from garf.io.writers import file_writer

try:
  import pandas as pd
except ImportError as e:
  raise ImportError(
    'Please install garf-io with Excel support - `pip install garf-io[excel]`'
  ) from e

logger = logging.getLogger(__name__)


class ExcelWriter(file_writer.FileWriter):
  """Writes Garf Report to Excel files.

  Attributes:
    destination_folder: Destination where Excel files are stored.
    file: Optional file to write report(s) to.
  """

  def __init__(
    self,
    destination_folder: Union[
      str, os.PathLike[str], pathlib.Path
    ] = pathlib.Path.cwd(),
    file: Optional[Union[str, os.PathLike[str], pathlib.Path]] = None,
    **kwargs,
  ) -> None:
    """Initializes ExcelWriter based on a destination_folder and file name.

    Args:
      destination_folder: Destination where Excel files are stored.
      file: Optional file to write report(s) to.
      kwargs: Optional keyword arguments to initialize writer.
    """
    self.file = file
    super().__init__(destination_folder=destination_folder, **kwargs)

  def __str__(self):
    return f'[Excel] - data are saved to {self.destination_folder} destination_folder.'

  @tracer.start_as_current_span('excel.write')
  def write(self, report: garf_report.GarfReport, destination: str) -> str:
    """Writes Garf report to a Excel sheet.

    Args:
      report: Garf report.
      destination: Sheet name report should be written to.

    Returns:
      Full path where data are written.
    """
    report = self.format_for_write(report)

    if not self.file:
      destination = formatter.format_extension(
        destination, new_extension='.xlsx'
      )
      output_path = os.path.join(self.destination_folder, destination)
      sheet_name = 'garf'
    else:
      output_path = os.path.join(self.destination_folder, self.file)
      sheet_name = destination
    self.create_dir()
    logger.debug(
      'Writing %d rows of data to file %s, sheet %s',
      len(report),
      output_path,
      destination,
    )
    with (
      smart_open.open(
        output_path,
        mode='wb',
      ) as f,
      pd.ExcelWriter(f, engine='openpyxl') as writer,
    ):
      report.to_pandas().to_excel(writer, sheet_name=sheet_name, index=False)
    logger.debug('Writing to %s is completed', output_path)
    return f'[Excel] - at {output_path}'



# Copyright 2026 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module for writing data to Google Sheets."""

from __future__ import annotations

from google.auth import exceptions as auth_exceptions

try:
  import gspread
except ImportError as e:
  raise ImportError(
    'Please install garf-io with sheets support - `pip install garf-io[sheets]`'
  ) from e

import logging
import pathlib
import uuid

from garf.core import report as garf_report
from garf.io import exceptions, formatter
from garf.io.telemetry import tracer
from garf.io.writers.abs_writer import AbsWriter
from opentelemetry import trace
from typing_extensions import override

logger = logging.getLogger(__name__)


class SheetWriterError(exceptions.GarfIoError):
  """SheetWriterError specific errors."""


class SheetWriter(AbsWriter):
  """Responsible for writing reports to Google Sheets."""

  def __init__(
    self,
    share_with: str | list[str] | None = None,
    credentials_file: str | None = None,
    spreadsheet_url: str | None = None,
    is_append: bool = False,
    **kwargs: str,
  ) -> None:
    """Initializes the SheetWriter to write reports to Google Sheets.

    Args:
      share_with: Email address to share the spreadsheet.
      credentials_file: Path to the service account credentials file.
      spreadsheet_url: URL of the Google Sheets spreadsheet.
      is_append: Whether you want to append data to the spreadsheet.
      spreadsheet: Initialized and shared spreadsheet.
    """
    super().__init__(**kwargs)
    if isinstance(share_with, str):
      share_with = share_with.split(',')
    elif share_with is None:
      share_with = []
    self.share_with = share_with
    self.credentials_file = credentials_file
    self.spreadsheet_url = spreadsheet_url
    self.is_append = is_append
    self._client = None

  @override
  @tracer.start_as_current_span('sheets.write')
  def write(
    self,
    report: garf_report.GarfReport,
    destination: str = f'Report {uuid.uuid4().hex}',
  ) -> str:
    report = self.format_for_write(report)
    if not destination:
      destination = f'Report {uuid.uuid4().hex}'
    destination = formatter.format_extension(destination)
    num_data_rows = len(report) + 1
    try:
      sheet = self.spreadsheet.worksheet(destination)
    except gspread.exceptions.WorksheetNotFound:
      sheet = self.spreadsheet.add_worksheet(
        destination, rows=num_data_rows, cols=len(report.column_names)
      )
    if not self.is_append:
      sheet.clear()
      self._add_rows_if_needed(num_data_rows, sheet)
      sheet.append_rows(
        [report.column_names] + report.results, value_input_option='RAW'
      )
    else:
      self._add_rows_if_needed(num_data_rows, sheet)
      sheet.append_rows(report.results, value_input_option='RAW')

    success_msg = f'Report is saved to {sheet.url}'
    logger.info(success_msg)
    return success_msg

  @property
  @tracer.start_as_current_span('sheets.create_client')
  def client(self) -> gspread.Client:
    span = trace.get_current_span()
    if self._client:
      return self._client
    config_dir = pathlib.Path.home() / '.config/gspread'
    if not self.credentials_file:
      if (credentials_file := config_dir / 'credentials.json').is_file():
        self._client = gspread.oauth(credentials_filename=credentials_file)
        span.set_attribute('sheets.auth_mode', 'oauth')
      elif (credentials_file := config_dir / 'service_account.json').is_file():
        self._client = self._init_service_account(credentials_file)
        span.set_attribute('sheets.auth_mode', 'service_account')
      else:
        raise SheetWriterError(
          'Failed to find either service_accounts.json or '
          'credentials.json files.'
          'Provide path to service account via `credentials_file` option'
        )
    else:
      try:
        self._client = self._init_service_account(self.credential_file)
      except auth_exceptions.MalformedError:
        self._client = gspread.oauth(credentials_filename=self.credentials_file)
    return self._client

  def _init_service_account(
    self, credentials_file: str | pathlib.Path
  ) -> gspread.Client:
    client = gspread.service_account(filename=credentials_file)
    if not self.spreadsheet_url:
      raise SheetWriterError(
        'Provide `spreadsheet_url` parameter when working with '
        'service account authentication.'
      )
    return client

  @tracer.start_as_current_span('sheets.create_or_get_spreadsheet')
  def create_or_get_spreadsheet(self) -> gspread.spreadsheet.Spreadsheet:
    span = trace.get_current_span()
    if not self.spreadsheet_url:
      spreadsheet = self.client.create(title=f'Garf {uuid.uuid4().hex}')
      span.set_attribute('sheets.is_new_spreadsheet', True)
      if not self.share_with:
        raise SheetWriterError('Provide your email in `share_with` parameter')
    else:
      spreadsheet = self.client.open_by_url(self.spreadsheet_url)
    if self.share_with:
      for email_address in self.share_with:
        spreadsheet.share(
          email_address=email_address,
          perm_type='user',
          role='writer',
          notify=False,
        )
    self.spreadsheet = spreadsheet
    return self.spreadsheet

  def _add_rows_if_needed(
    self, num_data_rows: int, sheet: gspread.worksheet.Worksheet
  ) -> None:
    num_sheet_rows = len(sheet.get_all_values())
    if num_data_rows > num_sheet_rows:
      num_rows_to_add = num_data_rows - num_sheet_rows
      sheet.add_rows(num_rows_to_add)

  def __str__(self) -> str:
    return f'[Sheet] - data are saved to {self.spreadsheet_url}.'
